class Axis {
  
  int x, y, height, graduationOutcrop, textBackgroundOpacityPercentage, grabHandleDiameter, grabHandleGap;
  String label, type;
  float min, max, graduationInterval;
  String[] values;
  color colour, textBackgroundColour;
  boolean baseLow, grabbed;
  
  Axis(String label, String type) {
     this.label = label;
     this.type = type;
     config();
     grabbed = false;
  }
  
  void config() {
    graduationOutcrop = 4;
    textBackgroundOpacityPercentage = 70;
    grabHandleDiameter = 8;
    grabHandleGap = 15;
  }
  
  void setOrder(String order) {
    if (order.equals("baseLow")) {
      baseLow = true;
    } else {
      baseLow = false;
    }
  }
  
  void setHeight(int height) {
    this.height = height;
    if (type.equals("string")) graduationInterval = (float) height / (float) (values.length - 1);
  }
  
  int getValueY(float value) {
    if (baseLow) {
      return (int) ceil(map(value, min, max, y + height, y));
    } else {
      return (int) ceil(map(value, min, max, y, y + height));
    }
  }
  
  int getValueIndex(String value) {
    int valueIndex = 0;
    for (int i = 0; i < values.length; i++) {
      if (value.equals(values[i])) {
        valueIndex = i;
      } 
    }
    return valueIndex;
  }
  
  int getValueY(String value) {
    return y + (int) (getValueIndex(value) * graduationInterval);
  }
  
  float getValueAt(int yValue) {
    if (type.equals("string")) {
      return 0;
    } else {
      if (baseLow) {
        return map(yValue, y, y + height, max, min);
      } else {
        return map(yValue, y, y + height, min, max);
      }
    }
  }
  
  void draw() {
    stroke(colour);
    strokeWeight(1);
    line(x, y, x, y + height);
    line(x - 5, y, x + 5, y);
    line(x - 5, y + height, x + 5, y + height);
    noStroke();
    fill(colour);
    textFont(sansSerif);
    textSize(10.5);
    textAlign(CENTER);
    text(label, x, y + height + 25);
  }
  
  void drawGraduations() {
    stroke(colour);
    strokeWeight(1);
    for (int i = 1; i < (values.length - 1); i++) {
      line(x - graduationOutcrop, y + (graduationInterval * i), x + graduationOutcrop, y + (graduationInterval * i));
    }
  }
  
  boolean hover() {
    return (mouseX >= x - graduationOutcrop && mouseX <= x + graduationOutcrop && mouseY >= y && mouseY <= y + height);
  }
  
  void drawValueLabels() {
    if (type.equals("string")) {
      for (int i = 0; i < values.length; i++) {
        if (mouseY >= y + (graduationInterval * i) - graduationOutcrop && mouseY <= y + (graduationInterval * i) + graduationOutcrop) {
          drawValueLabel(i, CENTER);
        }
      }
    } else {
      drawValueLabel(mouseY, CENTER);
    }
  }
  
  void drawValueLabel(int yPosition, int alignment) {
    textFont(sansSerif);
    textSize(9);
    textAlign(LEFT, TOP);
    int lineWidth = 1;
    float lineHeight = textAscent() + textDescent();
    int labelPadding = 4;
    String valueLabel = "";
    int rectX = x + graduationOutcrop + labelPadding;
    int rectY = 0;
    if (type.equals("string")) {
      valueLabel = values[yPosition];
      if (alignment == TOP) {
        rectY = (int) (y + (graduationInterval * yPosition) - (labelPadding * 2) - lineHeight);
      } else if (alignment == BOTTOM) {
        rectY = (int) (y + (graduationInterval * yPosition));
      } else {
        rectY = (int) (y + (graduationInterval * yPosition) - labelPadding - (lineHeight / 2)); 
      }
    } else { // seconds, float or int
      stroke(colour);
      strokeWeight(1);
      line(x - graduationOutcrop, yPosition, x + graduationOutcrop, yPosition);
      if (type.equals("seconds")) {
        int seconds;
        if (baseLow) {
          seconds = round(map(yPosition, y, y + height, max, min));
        } else {
          seconds = round(map(yPosition, y, y + height, min, max));
        }
        int hours = (int) floor(seconds / 3600);
        int minutes = (int) floor((seconds - (hours * 3600)) / 60);
        if (hours == 24) {
          hours = 23;
          minutes = 59;
        }
        if (hours < 10) valueLabel += "0";
        valueLabel += hours+":";
        if (minutes < 10) valueLabel += "0";
        valueLabel += minutes;
      } else { // int or float
        float value;
        if (baseLow) {
          value = map(yPosition, y, y + height, max, min);
        } else {
          value = map(yPosition, y, y + height, min, max);
        }
        if (type.equals("float")) {
          valueLabel = Float.toString(value);
        } else { // int
          valueLabel = Integer.toString((int) round(value));
        }
      }
      if (alignment == TOP) {
        rectY = (int) (yPosition - (labelPadding * 2) - lineHeight);
      } else if (alignment == BOTTOM) {
        rectY = yPosition;
      } else {
        rectY = (int) (yPosition - labelPadding - (lineHeight / 2));
      }
    }
    fill(textBackgroundColour, map(textBackgroundOpacityPercentage, 0, 100, 0, 255));
    noStroke();
    rect(rectX, rectY, textWidth(valueLabel) + (labelPadding * 2), lineHeight + (labelPadding * 2));
    fill(colour);
    text(valueLabel, rectX + labelPadding, rectY + labelPadding);
  }
  
  void drawGrabHandle() {
    stroke(colour);
    line(x, y - grabHandleGap, x, y);
    fill(colour);
    ellipse(x, y - grabHandleGap, grabHandleDiameter, grabHandleDiameter);
  }
  
  boolean grabHandleHover() {
    return (dist(mouseX, mouseY, x, y - grabHandleGap) <= (grabHandleDiameter / 2));
  }
}
