class Situation {
  
  boolean display, edit, viewValues;
  String label;
  color colour;
  Map axisStringValues; // maps axis label to an ArrayList of axis values which form part of this situation
  Map axisFloatValues; // as axisStringValues, but for non-string values (seconds and ints can also be stored as floats); stores an ArrayList of float[]
  Map axisYPoints; // maps axis label to an ArrayList of int arrays of the y points for drawing the situation
  Map axisRangeBars;
  
  Situation(String[] data, Map axes) {
    String[] colourComponents = data[0].split(",");
    colour = color(int(colourComponents[0]), int(colourComponents[1]), int(colourComponents[2]));
    label = data[1];
    axisStringValues = new HashMap();
    axisFloatValues = new HashMap();
    for (int i = 2; i < data.length; i++) {
      String[] attributes = split(data[i], ",");
      Axis axis = (Axis) axes.get(attributes[0]);
      ArrayList values = new ArrayList();
      for (int j = 1; j < attributes.length; j++) {
        if (axis.type.equals("string")) {
          values.add(attributes[j]);
        } else { // seconds, int or float
          String[] parts = attributes[j].split("-");
          if (parts.length == 2) { // range
            values.add(new float[] {float(parts[0]), float(parts[1])});
          } else { // single value
            values.add(new float[] {float(attributes[j])});
          }
        }
      }
      addAxisValues(axis.type, axis.label, values);
    }
    axisYPoints = new HashMap();
    axisRangeBars = new HashMap();
    display = false;
    edit = false;
    viewValues = false;
  }
  
  Situation(String label, Map axes, String[] axisOrder) {
    colour = color(255, 51, 51);
    this.label = label;
    axisStringValues = new HashMap();
    axisFloatValues = new HashMap();
    for (int i = 0; i < axisOrder.length; i++) {
      Axis axis = (Axis) axes.get(axisOrder[i]);
      ArrayList values = new ArrayList();
      addAxisValues(axis.type, axis.label, values);
    }
    axisYPoints = new HashMap();
    axisRangeBars = new HashMap();
    display = false;
    edit = false;
    viewValues = false;
  }
  
  void setColour(color colour, String[] axisOrder, Map axes) {
    this.colour = colour;
    for (int i = 0; i < axisOrder.length; i++) {
      Axis axis = (Axis) axes.get(axisOrder[i]);
      if (!axis.type.equals("string")) {
        ArrayList rangeBarsList = (ArrayList) axisRangeBars.get(axis.label);
        for (int j = 0; j < rangeBarsList.size(); j++) {
          ((RangeBar) rangeBarsList.get(j)).colour = colour;
        }
      }
    }
  }
  
  boolean containsValue(String axisLabel, String value) {
    return ((ArrayList) axisStringValues.get(axisLabel)).contains(value);
  }
  
  void addAxisValues(String type, String axisLabel, ArrayList values) {
    if (type.equals("string")) {
      axisStringValues.put(axisLabel, values);
    } else { // seconds, int or float
      axisFloatValues.put(axisLabel, values);
    }
  }
  
  void addAxisValue(String axisLabel, String value) {
    ArrayList values = (ArrayList) axisStringValues.get(axisLabel);
    if (!values.contains(value)) {
      values.add(value);
    }
  }
  
  void addAxisValue(String axisLabel, float value) {
  }
  
  void removeAxisValue(String axisLabel, String value) {
    ArrayList values = (ArrayList) axisStringValues.get(axisLabel);
    if (values.contains(value)) {
      values.remove(values.indexOf(value));
    }
  }
  
  void convertRangeBarsToValues(String[] axisOrder, Map axes) {
    for (int i = 0; i < axisOrder.length; i++) {
      Axis axis = (Axis) axes.get(axisOrder[i]);
      if (!axis.type.equals("string")) {
        ArrayList valuesRanges = new ArrayList();
        ArrayList rangeBarsList = (ArrayList) axisRangeBars.get(axis.label);
        int startY = 0;
        int endY = 0;
        for (int j = 0; j < rangeBarsList.size(); j++) {
          RangeBar rangeBar = (RangeBar) rangeBarsList.get(j);
          if (j == 0) { // first
            startY = rangeBar.startY;
            endY = rangeBar.endY;
            if (rangeBarsList.size() == 1) { // only item
              if (startY == endY) {
                valuesRanges.add(new float[] {axis.getValueAt(startY)});
              } else {
                valuesRanges.add(new float[] {axis.getValueAt(startY), axis.getValueAt(endY)});
              }
            }
          } else { // not first
            if (rangeBar.startY != endY) {
              if (startY == endY) {
                valuesRanges.add(new float[] {axis.getValueAt(startY)});
              } else {
                valuesRanges.add(new float[] {axis.getValueAt(startY), axis.getValueAt(endY)});
              }
              startY = rangeBar.startY;
              endY = rangeBar.endY;
            } else {
              endY = rangeBar.endY;
            }
            if (j == rangeBarsList.size() - 1) { // last
              if (startY == endY) {
                valuesRanges.add(new float[] {axis.getValueAt(startY)});
              } else {
                valuesRanges.add(new float[] {axis.getValueAt(startY), axis.getValueAt(endY)});
              }
            }
          }
        }
        axisFloatValues.put(axis.label, valuesRanges);
      }
    }
  }
  
  void calculatePoints(String[] axisOrder, Map axes) {
    for (int i = 0; i < axisOrder.length; i++) {
      Axis axis = (Axis) axes.get(axisOrder[i]);
      ArrayList yPoints = new ArrayList(); // there isn't a y point on the axis for every value when drawing a situation becuase there may be ranges (in which case only the start and end y point should be present)
      if (axis.type.equals("string")) {
        ArrayList values = (ArrayList) axisStringValues.get(axis.label); // axis values which are part of this situation
        int[] valueIndexes = new int[values.size()]; // for the axis indexes of values which are part of this situation
        for (int j = 0; j < values.size(); j++) {
          valueIndexes[j] = axis.getValueIndex((String) values.get(j));
        }
        // work out the ranges in the axis indexes
        Arrays.sort(valueIndexes); // sort axis value indexes sequentially
        boolean inRange = false;
        int rangeStart = -1;
        for (int j = 0; j < valueIndexes.length; j++) { // for each value index on the axis that forms part of this situation (remember we have sorted these numerically, low -> high starting from the top of the axis)
          if (j < (valueIndexes.length - 1) && // not at the last item
              valueIndexes[j + 1] - valueIndexes[j] == 1) { // we are in a range with the next item
            if (!inRange) {
              rangeStart = j;
              inRange = true;
            }
          } else { // the next item is not part of range (or there is no next item)
            if (inRange) {
              inRange = false;
              yPoints.add(new int[] {axis.getValueY(axis.values[valueIndexes[rangeStart]]), axis.getValueY(axis.values[valueIndexes[j]])});
            } else { // lone value
              yPoints.add(new int[] {axis.getValueY(axis.values[valueIndexes[j]])});
            }
          }
        }
        axisYPoints.put(axis.label, yPoints);
      } else { // seconds, int or float axis
        ArrayList valuesRanges = (ArrayList) axisFloatValues.get(axis.label);
        for (int j = 0; j < valuesRanges.size(); j++) {
          float[] values = (float[]) valuesRanges.get(j);
          if (values.length == 1) {
            yPoints.add(new int[] {axis.getValueY(values[0])});
          } else {
            yPoints.add(new int[] {axis.getValueY(values[0]), axis.getValueY(values[1])});
          }
        }
      }
      axisYPoints.put(axis.label, yPoints);
    }
  }
  
  void draw(String[] axisOrder, Map axes, int percentageOpactity) {
    for (int i = 0; i < (axisOrder.length - 1); i++) {
      Axis leftAxis = (Axis) axes.get(axisOrder[i]);
      Axis rightAxis = (Axis) axes.get(axisOrder[i + 1]);
      ArrayList leftAxisYPoints = (ArrayList) axisYPoints.get(leftAxis.label);
      ArrayList rightAxisYPoints = (ArrayList) axisYPoints.get(rightAxis.label);
      for (int j = 0; j < leftAxisYPoints.size(); j++) {
        int[] leftAxisYPointsArray = (int[]) leftAxisYPoints.get(j);
        for (int k = 0; k < rightAxisYPoints.size(); k++) {
          int[] rightAxisYPointsArray = (int[]) rightAxisYPoints.get(k);
          if (leftAxisYPointsArray.length == 1 && rightAxisYPointsArray.length == 1) {
            strokeWeight(2);
            stroke(colour, map(percentageOpactity, 0, 100, 0, 255));
            line(leftAxis.x, leftAxisYPointsArray[0], rightAxis.x, rightAxisYPointsArray[0]);
          } else {
            noStroke();
            fill(colour, map(percentageOpactity, 0, 100, 0, 255));
            beginShape();
            vertex(leftAxis.x, leftAxisYPointsArray[0]);
            vertex(rightAxis.x, rightAxisYPointsArray[0]);
            if (rightAxisYPointsArray.length == 2) {
              vertex(rightAxis.x, rightAxisYPointsArray[1]);
            }
            if (leftAxisYPointsArray.length == 2) {
              vertex(leftAxis.x, leftAxisYPointsArray[1]);
            }
            endShape();
          }
        }
      }
    }
  }
  
  void generateRangeBars(String[] axisOrder, Map axes) {
    for (int i = 0; i < axisOrder.length; i++) {
      Axis axis = (Axis) axes.get(axisOrder[i]);
      if (!axis.type.equals("string")) {
        ArrayList yPointRanges = (ArrayList) axisYPoints.get(axis.label);
        ArrayList rangeBarsList = new ArrayList();
        for (int j = 0; j < yPointRanges.size(); j++) {
          int minY = axis.y;
          int maxY = axis.y + axis.height;
          if (j > 0) {
            int[] previousYPoints = (int[]) yPointRanges.get(j - 1);
            if (previousYPoints.length == 2) {
              minY = previousYPoints[1];
            } else {
              minY = previousYPoints[0];
            }
          }
          if (j < yPointRanges.size() - 1) {
            int[] nextYPoints = (int[]) yPointRanges.get(j + 1);
            maxY = nextYPoints[0];
          }
          int[] yPoints = (int[]) yPointRanges.get(j);
          if (yPoints.length == 2) {
            rangeBarsList.add(new RangeBar(axis.x, yPoints[0], yPoints[1], minY, maxY, colour, axis.colour));
          } else {
            rangeBarsList.add(new RangeBar(axis.x, yPoints[0], yPoints[0], minY, maxY, colour, axis.colour));
          }
        }
        axisRangeBars.put(axis.label, rangeBarsList);
      }
    }
  }
  
  void drawValuesViewer(Point viewBox, int viewBoxWidth) {
    textFont(sansSerif);
    textSize(9);
    textAlign(LEFT, TOP);
    float maxTextWidth = 0;
    float lineHeight = textAscent() + textDescent();
    String[] valuesStrings = new String[axisOrder.length];
    for (int i = 0; i < axisOrder.length; i++) {
      Axis axis = (Axis) axes.get(axisOrder[i]);
      valuesStrings[i] = axis.label+": ";
      if (axis.type.equals("string")) {
        ArrayList stringValues = (ArrayList) axisStringValues.get(axis.label);
        for (int j = 0; j < stringValues.size(); j++) {
          if (j > 0) valuesStrings[i]+="|";
          valuesStrings[i] += (String) stringValues.get(j);
        }
      } else { // seconds, int or float
        ArrayList valueRanges = (ArrayList) axisFloatValues.get(axis.label);
        for (int j = 0; j < valueRanges.size(); j++) {
          if (j > 0) valuesStrings[i]+="|";
          float[] values = (float[]) valueRanges.get(j);
          if (axis.type.equals("float")) {
            valuesStrings[i] += values[0];
            if (values.length == 2) valuesStrings[i] += "-"+values[1];
          } else if (axis.type.equals("int")) {
            valuesStrings[i] += int(values[0]);
            if (values.length == 2) valuesStrings[i] += "-"+int(values[1]); 
          } else if (axis.type.equals("seconds")) {
            int hours, minutes;
            hours = (int) floor(values[0] / 3600);
            minutes = (int) floor((values[0] - (hours * 3600)) / 60);
            if (hours == 24) {
              hours = 23;
              minutes = 59;
            }
            if (hours < 10) valuesStrings[i] += "0";
            valuesStrings[i] += hours+":";
            if (minutes < 10) valuesStrings[i] += "0";
            valuesStrings[i] += minutes;
            if (values.length == 2) {
              valuesStrings[i] += "-";
              hours = (int) floor(values[1] / 3600);
              minutes = (int) floor((values[1] - (hours * 3600)) / 60);
              if (hours == 24) {
                hours = 23;
                minutes = 59;
              }
              if (hours < 10) valuesStrings[i] += "0";
              valuesStrings[i] += hours+":";
              if (minutes < 10) valuesStrings[i] += "0";
              valuesStrings[i] += minutes;
            }
          }
        }
      }
      if (textWidth(valuesStrings[i]) > maxTextWidth) maxTextWidth = textWidth(valuesStrings[i]);
    }
    int rectX = viewBox.x - ((int) maxTextWidth) - 20 - 15;
    int rectWidth = ((int) maxTextWidth) + 20;
    fill(255);
    noStroke();
    rect(rectX + rectWidth, viewBox.y, 15, viewBoxWidth);
    strokeWeight(1);
    stroke(0);
    rect(rectX, viewBox.y, rectWidth, (lineHeight * ((valuesStrings.length * 2) + 1)));
    fill(0);
    for (int i = 0; i < valuesStrings.length; i++) {
      text(valuesStrings[i], rectX + 10, viewBox.y + (((i * 2) + 1) * lineHeight));
    }
  }
  
  String toString(int numDimensions) {
    String[] valuesStrings = new String[numDimensions + 2];
    valuesStrings[0] = int(red(colour))+","+int(green(colour))+","+int(blue(colour));
    valuesStrings[1] = label;
    for (int i = 0; i < axisOrder.length; i++) {
      Axis axis = (Axis) axes.get(axisOrder[i]);
      valuesStrings[i+2] = axis.label;
      if (axis.type.equals("string")) {
        ArrayList stringValues = (ArrayList) axisStringValues.get(axis.label);
        for (int j = 0; j < stringValues.size(); j++) {
          valuesStrings[i+2] += ","+(String) stringValues.get(j);
        }
      } else { // seconds, int or float
        ArrayList valueRanges = (ArrayList) axisFloatValues.get(axis.label);
        for (int j = 0; j < valueRanges.size(); j++) {
          float[] values = (float[]) valueRanges.get(j);
          if (axis.type.equals("float")) {
            valuesStrings[i+2] += ","+values[0];
            if (values.length == 2) valuesStrings[i+2] += "-"+values[1];
          } else if (axis.type.equals("int") || axis.type.equals("seconds")) {
            valuesStrings[i+2] += ","+int(values[0]);
            if (values.length == 2) valuesStrings[i+2] += "-"+int(values[1]);
          }
        }
      }
    }
    String valuesString = "";
    for (int i = 0; i < valuesStrings.length; i++) {
      if (i > 0) valuesString += TAB;
      valuesString += valuesStrings[i];
    }
    return valuesString;
  }
}
