int startWidth, startHeight, dataCaseOpacity, situationOpacity, numDimensions, numDataCases, cursorType, colourBoxWidth, displayBoxWidth, editBoxWidth, viewBoxWidth, hoveredColourIndex, addBoxX, addBoxY, newSituationNameX, newSituationNameY;
String title, colouringSituationLabel;
StringBuffer newSituationName;
boolean updateBuffer, brushing, draggingAxis, editingSituation, rangeBarResizing, selectingColour, viewingSituationValues, addingSituation;
long lastWindowResizeTime;
int[] margin, pcvPadding;
int[][] colourChoices;
color backgroundColour, axisColour, dataCasesColour, exploreBrushColour, brushColour;
PImage buffer, editIcon, editIconDisabled, viewIcon, viewIconDisabled, addIcon, addIconDisabled;
PFont sansSerif;
Map axes, situations, floatAxesUniqueValues, situationDisplayBoxes, situationColourBoxes, situationEditBoxes, situationViewBoxes;
String[] axisOrder;
DataCase[] dataCases;
Point brushStartPoint, colourSelector;

void setup() {
  loadParams();
  size(startWidth, startHeight);
  margin = new int[4];
  margin[0] = 15; // top
  margin[1] = 5; // right
  margin[2] = 20; // bottom
  margin[3] = 15; // left
  pcvPadding = new int[4];
  pcvPadding[0] = 50; // top
  pcvPadding[1] = 250; // right
  pcvPadding[2] = 25; // bottom
  pcvPadding[3] = 40; // left
  backgroundColour = color(0, 0, 0);
  axisColour = color(255, 255, 255);
  dataCasesColour = color(0, 160, 204);
  exploreBrushColour = color(255, 0, 0);
  brushColour = exploreBrushColour;
  colourChoices = new int[256][3];
  loadColourChoices();
  colourBoxWidth = 10;
  displayBoxWidth = 10;
  editBoxWidth = 11;
  viewBoxWidth = 11;
  sansSerif = createFont("Sans Serif", 10);
  // handle window resize events
  frame.setResizable(true);
  frame.addComponentListener(
    new ComponentAdapter() {
      public void componentResized(ComponentEvent e) {
        if (e.getSource() == frame && millis() > 500) {
          // componentResized events are fired continously while the window is being resized
          lastWindowResizeTime = millis();
          // see draw method for remainder of window resize handling implementation
        }
      }
    } 
  );
  loadDataCases();
  loadSituations();
  for (int i = 0; i < axisOrder.length; i++) {
    Axis axis = (Axis) axes.get(axisOrder[i]);
    axis.colour = axisColour;
    axis.textBackgroundColour = backgroundColour;
  }
  situationDisplayBoxes = new HashMap();
  situationColourBoxes = new HashMap();
  situationEditBoxes = new HashMap();
  situationViewBoxes = new HashMap();
  colourSelector = new Point();
  loadIcons();
  windowResized(); // calculate initial pixel positions before first draw call
  brushing = false;
  draggingAxis = false;
  editingSituation = false;
  rangeBarResizing = false;
  selectingColour = false;
  viewingSituationValues = false;
  addingSituation = false;
  newSituationName = new StringBuffer();
}

void loadParams() {
  Table paramsTable = new Table("params.tsv", false); // read all data from file
  for (int i = 0; i < paramsTable.data.length; i++) {
    if (paramsTable.data[i][0].equals("startWidth")) {
      startWidth = int(paramsTable.data[i][1]);
    } else if (paramsTable.data[i][0].equals("startHeight")) {
      startHeight = int(paramsTable.data[i][1]);
    } else if (paramsTable.data[i][0].equals("title")) {
      title = paramsTable.data[i][1];
    } else if (paramsTable.data[i][0].equals("dataCaseOpacity")) {
      dataCaseOpacity = int(paramsTable.data[i][1]);
    } else if (paramsTable.data[i][0].equals("situationOpacity")) {
      situationOpacity = int(paramsTable.data[i][1]);
    }
  }
}

void loadColourChoices() {
  /*
  Colour layout taken from http://html-color-codes.com/
  Copyright 2000 VisiBone
  Arrangement by Bob Stein
  */
  Table colourSelectorTable = new Table("colourSelector.tsv", false);
  for (int i = 0; i < colourSelectorTable.data.length; i++) {
    colourChoices[i] = new int[] {int(colourSelectorTable.data[i][0]), int(colourSelectorTable.data[i][1]), int(colourSelectorTable.data[i][2])};
  }
}

void loadDataCases() {
  // find all the attribute files in the data directory
  java.io.File folder = new java.io.File(dataPath(""));
  java.io.FilenameFilter attributeFileFilter = new java.io.FilenameFilter() {
    boolean accept(File dir, String name) {
      return name.matches("attribute[0-9]+.tsv");
    }
  };
  String[] attributeFilenames = folder.list(attributeFileFilter);
  numDimensions = attributeFilenames.length;
  axes = new HashMap();
  axisOrder = new String[numDimensions];
  // load in meta data from attribute files
  for (int i = 0; i < attributeFilenames.length; i++) {
     Table attributeDataTable = new Table(attributeFilenames[i], true); // read only meta data from attribute file
     Axis axis = new Axis(attributeDataTable.name, attributeDataTable.type);
     if (attributeDataTable.type.equals("string")) {
       axis.values = attributeDataTable.possibleValues;
     } else { // seconds, int or float
       axis.min = attributeDataTable.min;
       axis.max = attributeDataTable.max;
       axis.setOrder(attributeDataTable.order);
     }
     axis.y = margin[0] + pcvPadding[0];
     axes.put(axis.label, axis);
     axisOrder[i] = axis.label;
     numDataCases = attributeDataTable.numCases;
  }
  dataCases = new DataCase[numDataCases];
  for (int i = 0; i < dataCases.length; i++) {
    dataCases[i] = new DataCase(numDimensions);
    dataCases[i].colour = dataCasesColour;
  }
  floatAxesUniqueValues = new HashMap();
  // load in case data from attribute files
  for (int i = 0; i < attributeFilenames.length; i++) {
    Table attributeDataTable = new Table(attributeFilenames[i], false); // read all data from attribute file
    Axis axis = (Axis) axes.get(attributeDataTable.name);
    if (axis.type.equals("string")) {
      for (int j = 0; j < dataCases.length; j++) {
        dataCases[j].setValue(axis.label, attributeDataTable.data[j][0]);
      }
    } else {
      ArrayList uniqueValues = new ArrayList();
      for (int j = 0; j < dataCases.length; j++) {
        dataCases[j].setValue(axis.label, attributeDataTable.data[j][0]);
        if (!uniqueValues.contains(attributeDataTable.data[j][0])) {
          uniqueValues.add(float(attributeDataTable.data[j][0]));
        }
      }
      floatAxesUniqueValues.put(axis.label, uniqueValues);
    }
  }
}

void loadSituations() {
  situations = new HashMap();
  Table situationsTable = new Table("situations.tsv", false); // read all data from file
  for (int i = 0; i < situationsTable.data.length; i++) {
    Situation situation = new Situation(situationsTable.data[i], axes);
    situations.put(situation.label, situation);
  }
}

void loadIcons() {
  editIcon = loadImage("icons/edit.png");
  editIconDisabled = loadImage("icons/edit_disabled.png");
  viewIcon = loadImage("icons/view.png");
  viewIconDisabled = loadImage("icons/view_disabled.png");
  addIcon = loadImage("icons/add.png");
  addIconDisabled = loadImage("icons/add_disabled.png");
}

void windowResized() {
  for (int i = 0; i < axisOrder.length; i++) {
    Axis axis = (Axis) axes.get(axisOrder[i]);
    axis.setHeight(getHeight() - margin[0] - margin[2] - pcvPadding[0] - pcvPadding[2]);
    axis.x = (int) floor(margin[3] + pcvPadding[3] + (i * ((getWidth() - margin[1] - margin[3] - pcvPadding[1] - pcvPadding[3]) / (axes.size() - 1))));
  }
  calculateDataCasePoints();
  calculateSituationPoints();
  updateBuffer = true;
}

void calculateDataCasePoints() {
  for (int i = 0; i < axisOrder.length; i++) {
    Axis axis = (Axis) axes.get(axisOrder[i]);
    for (int j = 0; j < dataCases.length; j++) {
      dataCases[j].points[i].x = axis.x;
      if (axis.type.equals("string")) {
        dataCases[j].points[i].y = axis.getValueY((String) dataCases[j].values.get(axis.label));
      } else { // seconds, int or float
        dataCases[j].points[i].y = axis.getValueY(float((String) dataCases[j].values.get(axis.label)));
      }
    }
  }
}

void calculateSituationPoints() {
  Set situationsLabels = situations.keySet();
  Iterator situationsLabelsIterator = situationsLabels.iterator();
  while (situationsLabelsIterator.hasNext()) {
    Situation situation = (Situation) situations.get((String) situationsLabelsIterator.next());
    situation.calculatePoints(axisOrder, axes);
    situation.generateRangeBars(axisOrder, axes);
  }
}

void draw() {
  checkForWindowResize();
  smooth();
  drawBufferContents();
  cursorType = ARROW; // default
  if (viewingSituationValues) {
    Set situationsLabels = situations.keySet();
    Iterator situationsLabelsIterator = situationsLabels.iterator();
    while (situationsLabelsIterator.hasNext()) {
      Situation situation = (Situation) situations.get((String) situationsLabelsIterator.next());
      Point viewBox = (Point) situationViewBoxes.get(situation.label);
      if (situation.viewValues) {
        if (mouseX >= viewBox.x && mouseX <= viewBox.x + viewBoxWidth && mouseY >= viewBox.y && mouseY <= viewBox.y + viewBoxWidth) cursorType = HAND;
        situation.drawValuesViewer(viewBox, viewBoxWidth);
      }
    }
  } else if (selectingColour) {
    drawColourSelector();
  } else if (brushing) {
    cursorType = CROSS;
    stroke(brushColour);
    strokeWeight(1);
    line(brushStartPoint.x, brushStartPoint.y, constrain(mouseX, margin[3] + pcvPadding[3], getWidth() - (margin[1] + pcvPadding[1])), constrain(mouseY, margin[0] + pcvPadding[0], getHeight() - (margin[2] + pcvPadding[2])));
  } else if (draggingAxis) {
    cursorType = HAND;
    for (int i = 0; i < axisOrder.length; i++) {
      Axis axis = (Axis) axes.get(axisOrder[i]);
      if (axis.grabHandleHover()) axis.drawGrabHandle();
      if (axis.grabbed) {
        axis.drawGrabHandle();
        stroke(axisColour);
        strokeWeight(1);
        Point targetPoint = new Point(constrain(mouseX, margin[3] + pcvPadding[3], getWidth() - (margin[1] + pcvPadding[1])), axis.y - axis.grabHandleGap);
        line(axis.x, axis.y - axis.grabHandleGap, targetPoint.x, targetPoint.y);
        fill(axisColour);
        ellipse(targetPoint.x, targetPoint.y, (axis.grabHandleDiameter / 2), (axis.grabHandleDiameter / 2));
      }
    }
  } else if (editingSituation) {
    Set situationsLabels = situations.keySet();
    Iterator situationsLabelsIterator = situationsLabels.iterator();
    while (situationsLabelsIterator.hasNext()) {
      Situation situation = (Situation) situations.get((String) situationsLabelsIterator.next());
      if (situation.edit) {
        Point editBox = (Point) situationEditBoxes.get(situation.label);
        if (mouseX >= editBox.x && mouseX <= editBox.x + editBoxWidth && mouseY >= editBox.y && mouseY <= editBox.y + editBoxWidth) {
          cursorType = HAND;
        }
        for (int i = 0; i < axisOrder.length; i++) {
          Axis axis = (Axis) axes.get(axisOrder[i]);
          if (axis.type.equals("string")) {
            if (!rangeBarResizing && axis.hover()) {
              for (int j = 0; j < axis.values.length; j++) {
                if (mouseY >= axis.y + (axis.graduationInterval * j) - axis.graduationOutcrop && mouseY <= axis.y + (axis.graduationInterval * j) + axis.graduationOutcrop) {
                  cursorType = HAND;
                  axis.drawValueLabel(j, CENTER);
                  strokeWeight(1);
                  stroke(axis.colour);
                  if (situation.containsValue(axis.label, axis.values[j])) {
                    fill(0, 255, 0);
                  } else {
                    fill(255, 0, 0);
                  }
                  ellipse(axis.x, axis.y + (axis.graduationInterval * j), axis.graduationOutcrop * 2, axis.graduationOutcrop * 2);
                }
              }
            }
          } else { // seconds, int or float
            boolean showValueLabels = true;
            ArrayList rangeBarsList = (ArrayList) situation.axisRangeBars.get(axis.label);
            for (int j = 0; j < rangeBarsList.size(); j++) {
              RangeBar rangeBar = (RangeBar) rangeBarsList.get(j);
              if (!rangeBar.selected) {
                rangeBar.draw(!rangeBarResizing);
                if (!rangeBarResizing && rangeBar.hover()) {
                  cursorType = HAND;
                  showValueLabels = false;
                  axis.drawValueLabel(rangeBar.startY, TOP);
                  axis.drawValueLabel(rangeBar.endY, BOTTOM);
                }
              }
            }
            for (int j = 0; j < rangeBarsList.size(); j++) {
              RangeBar rangeBar = (RangeBar) rangeBarsList.get(j);
              if (rangeBar.selected) {
                if (rangeBar.resizing) {
                  cursorType = HAND;
                  if (rangeBar.dragPoint == BOTTOM) {
                    rangeBar.endY = constrain(mouseY - rangeBar.cordLength, rangeBar.startY, rangeBar.maxY);
                  } else if (rangeBar.dragPoint == TOP) {
                    rangeBar.startY = constrain(mouseY + rangeBar.cordLength, rangeBar.minY, rangeBar.endY);
                  }
                }
                rangeBar.draw(!rangeBarResizing);
                rangeBar.drawEditingComponents();
                if (!rangeBar.resizing) {
                  rangeBar.drawDeleteBox();
                  if (rangeBar.topHandleHover() || rangeBar.deleteBoxHover() || rangeBar.bottomHandleHover()) {
                    cursorType = HAND;
                  }
                }
                showValueLabels = false;
                axis.drawValueLabel(rangeBar.startY, TOP);
                axis.drawValueLabel(rangeBar.endY, BOTTOM);
              }
            }
            if (showValueLabels && axis.hover()) {
              axis.drawValueLabels();
            }
          }
        }
      }
    }
  } else if (addingSituation) {
    strokeWeight(1);
    stroke(axisColour);
    noFill();
    textFont(sansSerif);
    textAlign(LEFT, TOP);
    textSize(10);
    text("Add: "+newSituationName, newSituationNameX, newSituationNameY);
  } else if (mouseX > getWidth() - pcvPadding[1]) { // in situations panel
    if (mouseX >= addBoxX && mouseX <= addBoxX + addIcon.width && mouseY >= addBoxY && mouseY <= addBoxY + addIcon.height) {
      cursorType = HAND;
    } else {
      Set situationsLabels = situations.keySet();
      Iterator situationsLabelsIterator = situationsLabels.iterator();
      while (situationsLabelsIterator.hasNext()) {
        Situation situation = (Situation) situations.get((String) situationsLabelsIterator.next());
        Point colourBox = (Point) situationColourBoxes.get(situation.label);
        if (mouseX >= colourBox.x && mouseX <= colourBox.x + colourBoxWidth && mouseY >= colourBox.y && mouseY <= colourBox.y + colourBoxWidth) cursorType = HAND;
        Point displayBox = (Point) situationDisplayBoxes.get(situation.label);
        if (mouseX >= displayBox.x && mouseX <= displayBox.x + displayBoxWidth && mouseY >= displayBox.y && mouseY <= displayBox.y + displayBoxWidth) cursorType = HAND;
        Point editBox = (Point) situationEditBoxes.get(situation.label);
        if (mouseX >= editBox.x && mouseX <= editBox.x + editBoxWidth && mouseY >= editBox.y && mouseY <= editBox.y + editBoxWidth) cursorType = HAND;
        Point viewBox = (Point) situationViewBoxes.get(situation.label);
        if (mouseX >= viewBox.x && mouseX <= viewBox.x + viewBoxWidth && mouseY >= viewBox.y && mouseY <= viewBox.y + viewBoxWidth) cursorType = HAND;
      }
    }
  } else if (mouseY < margin[0] + pcvPadding[0]) { // above PCV drawing area
    for (int i = 0; i < axisOrder.length; i++) {
      Axis axis = (Axis) axes.get(axisOrder[i]);
      if (axis.grabHandleHover()) {
        axis.drawGrabHandle();
        cursorType = HAND;
      }
    }
  } else {
    for (int i = 0; i < axisOrder.length; i++) {
      Axis axis = (Axis) axes.get(axisOrder[i]);
      if (axis.hover()) {
        axis.drawValueLabels();
      }
    }
  }
  if (!editingSituation) {
    int countSelectedDataCases = 0;
    for (int i = 0; i < dataCases.length; i++) { // for each data case
      if (dataCases[i].drawOrder > 0) countSelectedDataCases++;
    }
    if (countSelectedDataCases > 0) {
      fill(brushColour);
      textFont(sansSerif);
      textAlign(RIGHT, TOP);
      textSize(10);
      text("Press 'c' to clear selections", getWidth() - (margin[1] * 2), getHeight() - pcvPadding[2] - 5);
    }
  }
  cursor(cursorType);
}

void checkForWindowResize() {
  if (lastWindowResizeTime > 0 && (millis() - lastWindowResizeTime) > 500) { // rough way to tell if a window resize has just finished
    lastWindowResizeTime = 0;
    windowResized();
  }
}

void drawBufferContents() {
  if (updateBuffer) {
    updateBuffer = false;
    drawStaticContent();
    buffer = createImage(getWidth(), getHeight(), RGB);
    loadPixels();
    for (int i = 0; i < getWidth() * getHeight(); i++) {
      buffer.pixels[i] = pixels[i];
    }
  } else {
    image(buffer, 0, 0);
  }
}

void drawStaticContent() {
  background(backgroundColour);
  fill(axisColour);
  textFont(sansSerif);
  textAlign(CENTER, TOP);
  textSize(16);
  text(title, (getWidth() / 2), margin[0]);
  Arrays.sort(dataCases); // sorts by drawOrder int defined in object
  for (int i = 0; i < dataCases.length; i++) {
    if (dataCases[i].drawOrder > 0) {
      dataCases[i].draw(100);
    } else {
      dataCases[i].draw(dataCaseOpacity);
    }
  }
  Set situationsLabels = situations.keySet();
  Iterator situationsLabelsIterator = situationsLabels.iterator();
  while (situationsLabelsIterator.hasNext()) {
    Situation situation = (Situation) situations.get((String) situationsLabelsIterator.next());
    if (situation.display) {
      situation.draw(axisOrder, axes, situationOpacity);
    }
  }
  for (int i = 0; i < axisOrder.length; i++) {
    Axis axis = (Axis) axes.get(axisOrder[i]);
    axis.draw();
    if (axis.type.equals("string")) axis.drawGraduations();
  }
  drawSituationsPanel();
}

void drawSituationsPanel() {
  strokeWeight(1);
  stroke(axisColour);
  noFill();
  int rectX = getWidth() - (pcvPadding[1] + margin[1]) + margin[3];
  int rectY = margin[0] + pcvPadding[0];
  int rectWidth = pcvPadding[1] - margin[1] - margin[3];
  int rectHeight = getHeight() - (margin[0] + margin[2]) - (pcvPadding[0] + pcvPadding[2]);
  rect(rectX, rectY, rectWidth, rectHeight);
  addBoxX = rectX + rectWidth - addIcon.width;
  addBoxY = rectY + rectHeight - addIcon.height;
  if (addingSituation || editingSituation || viewingSituationValues) {
    drawAddIcon(addBoxX, addBoxY, true);
  } else {
    drawAddIcon(addBoxX, addBoxY, false);
  }
  textFont(sansSerif);
  textSize(13);
  textAlign(CENTER, TOP);
  text("Situations", rectX + (rectWidth / 2), rectY + 15);
  textAlign(LEFT, TOP);
  textSize(10);
  float lineHeight = textAscent() + textDescent();
  newSituationNameX = rectX + 7;
  newSituationNameY = int(rectY + rectHeight - lineHeight - 3);
  Set situationsLabels = situations.keySet();
  Iterator situationsLabelsIterator = situationsLabels.iterator();
  int i = 0;
  while (situationsLabelsIterator.hasNext()) {
    Situation situation = (Situation) situations.get((String) situationsLabelsIterator.next());
    int colourBoxX = rectX + colourBoxWidth;
    int colourBoxY = rectY + 65 + (int) (i * lineHeight) + (i * 15);
    fill(situation.colour);
    rect(colourBoxX, colourBoxY, colourBoxWidth, colourBoxWidth);
    situationColourBoxes.put(situation.label, new Point(colourBoxX, colourBoxY));
    noFill();
    int displayBoxX = colourBoxX + round(displayBoxWidth * 1.5);
    rect(displayBoxX, colourBoxY, displayBoxWidth, displayBoxWidth);
    if (situation.display) {
      line(displayBoxX, colourBoxY, displayBoxX + displayBoxWidth, colourBoxY + displayBoxWidth);
      line(displayBoxX, colourBoxY + displayBoxWidth, displayBoxX + displayBoxWidth, colourBoxY);
    }
    situationDisplayBoxes.put(situation.label, new Point(displayBoxX, colourBoxY));
    fill(axisColour);
    text(situation.label, displayBoxX + (displayBoxWidth * 2), colourBoxY);
    int viewIconX = rectX + rectWidth - viewBoxWidth - (viewBoxWidth / 2);
    if (editingSituation) {
      drawViewIcon(viewIconX, colourBoxY, true);
    } else {
      if (viewingSituationValues) {
        if (situation.viewValues) {
          drawViewIcon(viewIconX, colourBoxY, false);
        } else {
          drawViewIcon(viewIconX, colourBoxY, true);
        }
      } else {
        drawViewIcon(viewIconX, colourBoxY, false);
      }
    }
    situationViewBoxes.put(situation.label, new Point(viewIconX, colourBoxY));
    int editIconX = viewIconX - viewBoxWidth - (viewBoxWidth / 2);
    if (viewingSituationValues) {
      drawEditIcon(editIconX, colourBoxY, true);
    } else {
      if (editingSituation) {
        if (situation.edit) {
          drawEditIcon(editIconX, colourBoxY, false);
        } else {
          drawEditIcon(editIconX, colourBoxY, true);
        }
      } else {
        drawEditIcon(editIconX, colourBoxY, false);
      }
    }
    situationEditBoxes.put(situation.label, new Point(editIconX, colourBoxY));
    i++;
  }
}

void drawColourSelector() {
  strokeWeight(1);
  stroke(axisColour);
  fill(255);
  int rectX = colourSelector.x - (colourBoxWidth * 16) - colourBoxWidth;
  int rectY = colourSelector.y;
  rect(colourSelector.x - colourBoxWidth, rectY, colourBoxWidth, colourBoxWidth);
  stroke(0);
  rect(rectX, rectY, (colourBoxWidth * 16), (colourBoxWidth * 16));
  int row = 0;
  int col = 0;
  hoveredColourIndex = -1;
  for (int i = 0; i < colourChoices.length; i++) {
    if (i > 0 && i % 16 == 0) {
      row++;
      col = 0;
    }
    if (colourChoices[i][0] == -1) {
      line((rectX + (colourBoxWidth * col)), (rectY + (colourBoxWidth * row)), (rectX + (colourBoxWidth * col)) + colourBoxWidth, (rectY + (colourBoxWidth * row)) + colourBoxWidth);
    } else {
      fill(colourChoices[i][0], colourChoices[i][1], colourChoices[i][2]);
      rect((rectX + (colourBoxWidth * col)), (rectY + (colourBoxWidth * row)), colourBoxWidth, colourBoxWidth);
      if (mouseX >= (rectX + (colourBoxWidth * col)) && mouseX <= (rectX + (colourBoxWidth * col) + colourBoxWidth) && mouseY >= (rectY + (colourBoxWidth * row)) && mouseY <= (rectY + (colourBoxWidth * row) + colourBoxWidth)) {
        cursorType = HAND;
        hoveredColourIndex = i;
      }
    }
    col++;
  }
}

void drawEditIcon(int x, int y, boolean disabled) {
  if (disabled) {
    image(editIconDisabled, x, y);
  } else {
    image(editIcon, x, y);
  }
}

void drawViewIcon(int x, int y, boolean disabled) {
  if (disabled) {
    image(viewIconDisabled, x, y);
  } else {
    image(viewIcon, x, y);
  }
}

void drawAddIcon(int x, int y, boolean disabled) {
  if (disabled) {
    image(addIconDisabled, x, y);
  } else {
    image(addIcon, x, y);
  }
}

void mousePressed() {
  if (mouseButton == LEFT) {
    if (viewingSituationValues) {
      Set situationsLabels = situations.keySet();
      Iterator situationsLabelsIterator = situationsLabels.iterator();
      while (situationsLabelsIterator.hasNext()) {
        Situation situation = (Situation) situations.get((String) situationsLabelsIterator.next());
        Point viewBox = (Point) situationViewBoxes.get(situation.label);
        if (situation.viewValues && mouseX >= viewBox.x && mouseX <= viewBox.x + viewBoxWidth && mouseY >= viewBox.y && mouseY <= viewBox.y + viewBoxWidth) {
          situation.viewValues = false;
          viewingSituationValues = false;
          updateBuffer = true;
        }
      }
    } else if (selectingColour) {
      if (hoveredColourIndex >= 0) {
        Situation situation = (Situation) situations.get(colouringSituationLabel);
        situation.setColour(color(colourChoices[hoveredColourIndex][0], colourChoices[hoveredColourIndex][1], colourChoices[hoveredColourIndex][2]), axisOrder, axes);
        selectingColour = false;
        updateBuffer = true;
        saveSituations();
      }
    } else if (editingSituation) {
      Set situationsLabels = situations.keySet();
      Iterator situationsLabelsIterator = situationsLabels.iterator();
      while (situationsLabelsIterator.hasNext()) {
        Situation situation = (Situation) situations.get((String) situationsLabelsIterator.next());
        if (situation.edit) {
          Point editBox = (Point) situationEditBoxes.get(situation.label);
          if (mouseX >= editBox.x && mouseX <= editBox.x + editBoxWidth && mouseY >= editBox.y && mouseY <= editBox.y + editBoxWidth) {
            brushColour = exploreBrushColour;
            situation.edit = false;
            editingSituation = false;
            updateBuffer = true;
            saveSituations();
          }
          for (int i = 0; i < axisOrder.length; i++) {
            Axis axis = (Axis) axes.get(axisOrder[i]);
            if (axis.type.equals("string")) {
              for (int j = 0; j < axis.values.length; j++) {
                if (axis.hover() && mouseY >= axis.y + (axis.graduationInterval * j) - axis.graduationOutcrop && mouseY <= axis.y + (axis.graduationInterval * j) + axis.graduationOutcrop) {
                  if (situation.containsValue(axis.label, axis.values[j])) {
                    situation.removeAxisValue(axis.label, axis.values[j]);
                  } else {
                    situation.addAxisValue(axis.label, axis.values[j]);
                  }
                  situation.calculatePoints(axisOrder, axes);
                  updateBuffer = true;
                }
              }
            } else { // seconds, int or float
              ArrayList rangeBarsList = (ArrayList) situation.axisRangeBars.get(axis.label);
              int removeRangeBar = -1;
              for (int j = 0; j < rangeBarsList.size(); j++) {
                RangeBar rangeBar = (RangeBar) rangeBarsList.get(j);
                if (rangeBar.selected) {
                  if (rangeBar.topHandleHover()) {
                    rangeBarResizing = true;
                    rangeBar.resizing = true;
                    rangeBar.dragPoint = TOP;
                  } else if (rangeBar.bottomHandleHover()) {
                    rangeBarResizing = true;
                    rangeBar.resizing = true;
                    rangeBar.dragPoint = BOTTOM;
                  } else if (rangeBar.deleteBoxHover()) {
                    removeRangeBar = j;
                  }
                }
              }
              if (removeRangeBar >= 0) {
                rangeBarsList.remove(removeRangeBar);
                situation.convertRangeBarsToValues(axisOrder, axes);
                situation.calculatePoints(axisOrder, axes);
                situation.generateRangeBars(axisOrder, axes);
                updateBuffer = true;
              } else if (!rangeBarResizing) {
                for (int j = 0; j < rangeBarsList.size(); j++) {
                  RangeBar rangeBar = (RangeBar) rangeBarsList.get(j);
                  if (rangeBar.hover()) {
                    rangeBar.selected = true;
                  } else {
                    rangeBar.selected = false;
                  }
                }
              }
            }
          }
        }
      }
    } else if (mouseX > getWidth() - pcvPadding[1]) { // in situations panel
      if (!addingSituation && mouseX >= addBoxX && mouseX <= addBoxX + addIcon.width && mouseY >= addBoxY && mouseY <= addBoxY + addIcon.height) {
        addingSituation = true;
        updateBuffer = true;
      } else {
        Set situationsLabels = situations.keySet();
        Iterator situationsLabelsIterator = situationsLabels.iterator();
        while (situationsLabelsIterator.hasNext()) {
          Situation situation = (Situation) situations.get((String) situationsLabelsIterator.next());
          Point colourBox = (Point) situationColourBoxes.get(situation.label);
          if (mouseX >= colourBox.x && mouseX <= colourBox.x + colourBoxWidth && mouseY >= colourBox.y && mouseY <= colourBox.y + colourBoxWidth) {
            colourSelector.x = colourBox.x;
            colourSelector.y = colourBox.y;
            colouringSituationLabel = situation.label;
            selectingColour = true;
          }
          Point displayBox = (Point) situationDisplayBoxes.get(situation.label);
          if (mouseX >= displayBox.x && mouseX <= displayBox.x + displayBoxWidth && mouseY >= displayBox.y && mouseY <= displayBox.y + displayBoxWidth) {
            situation.display = !situation.display;
            updateBuffer = true;
          }
          Point viewBox = (Point) situationViewBoxes.get(situation.label);
          if (mouseX >= viewBox.x && mouseX <= viewBox.x + viewBoxWidth && mouseY >= viewBox.y && mouseY <= viewBox.y + viewBoxWidth) {
            situation.viewValues = true;
            viewingSituationValues = true;
            updateBuffer = true;
          }
          Point editBox = (Point) situationEditBoxes.get(situation.label);
          if (mouseX >= editBox.x && mouseX <= editBox.x + editBoxWidth && mouseY >= editBox.y && mouseY <= editBox.y + editBoxWidth) {
            brushColour = situation.colour;
            situation.display = true;
            situation.edit = true;
            editingSituation = true;
            updateBuffer = true;
          }
        }
      }
    } else if (mouseY < margin[0] + pcvPadding[0]) { // above PCV drawing area
      for (int i = 0; i < axisOrder.length; i++) {
        Axis axis = (Axis) axes.get(axisOrder[i]);
        if (axis.grabHandleHover()) {
          axis.grabbed = true;
          draggingAxis = true;
        }
      }
    }
  } else if (mouseButton == RIGHT) {
    if (mouseX >= margin[3] + pcvPadding[3] && mouseX <=  getWidth() - (margin[1] + pcvPadding[1]) && mouseY >= margin[0] + pcvPadding[0] && mouseY <= getHeight() - (margin[2] + pcvPadding[2])) { // inside PCV drawing area
      brushStartPoint = new Point(mouseX, mouseY);
      brushing = true;
    }
  }
}

void mouseReleased() {
  if (editingSituation) {
    Set situationsLabels = situations.keySet();
    Iterator situationsLabelsIterator = situationsLabels.iterator();
    while (situationsLabelsIterator.hasNext()) {
      Situation situation = (Situation) situations.get((String) situationsLabelsIterator.next());
      if (situation.edit) {
        for (int i = 0; i < axisOrder.length; i++) {
          Axis axis = (Axis) axes.get(axisOrder[i]);
          if (!axis.type.equals("string")) { // seconds, int or float
            if (rangeBarResizing) {
              rangeBarResizing = false;
              ArrayList rangeBarsList = (ArrayList) situation.axisRangeBars.get(axis.label);
              for (int j = 0; j < rangeBarsList.size(); j++) {
                RangeBar rangeBar = (RangeBar) rangeBarsList.get(j);
                if (rangeBar.resizing) {
                  rangeBar.resizing = false;
                  rangeBar.dragPoint = -1;
                  situation.convertRangeBarsToValues(axisOrder, axes);
                  situation.calculatePoints(axisOrder, axes);
                  situation.generateRangeBars(axisOrder, axes);
                  updateBuffer = true;
                }
              }
            }
          }
        }
        if (brushing) {
          brushing = false;
          Point brushEndPoint = new Point(mouseX, mouseY);
          ArrayList dataCasesSelections = new ArrayList();
          for (int i = 0; i < dataCases.length; i++) {
            if (dataCases[i].intersectedBy(brushStartPoint, brushEndPoint)) {
              dataCasesSelections.add((Integer) i);
            }
          }
          for (int i = 0; i < axisOrder.length; i++) {
            Axis axis = (Axis) axes.get(axisOrder[i]);
            if (axis.type.equals("string")) {
              for (int j = 0; j < dataCasesSelections.size(); j++) {
                situation.addAxisValue(axis.label, (String) dataCases[(Integer) dataCasesSelections.get(j)].values.get(axis.label));
              }
            } else { // seconds, int or float
              ArrayList uniqueValues = (ArrayList) floatAxesUniqueValues.get(axis.label);
              boolean[] selectedValues = new boolean[uniqueValues.size()];
              for (int j = 0; j < dataCasesSelections.size(); j++) {
                selectedValues[uniqueValues.indexOf((Object) float((String) dataCases[(Integer) dataCasesSelections.get(j)].values.get(axis.label)))] = true;
              }
              // work out the ranges for the brushed cases
              ArrayList yRanges = new ArrayList();
              boolean inRange = false;
              int rangeStart = -1;
              for (int j = 0; j < selectedValues.length; j++) {
                if (j < (selectedValues.length - 1) && selectedValues[j] && selectedValues[(j + 1)]) { // sequential
                  if (!inRange) {
                    rangeStart = j;
                    inRange = true;
                  }
                } else {
                  if (inRange) {
                    inRange = false;
                    yRanges.add(new int[] {axis.getValueY((Float) uniqueValues.get(rangeStart)), axis.getValueY((Float) uniqueValues.get(j))});
                  } else if (selectedValues[j]) { // lone value
                    yRanges.add(new int[] {axis.getValueY((Float) uniqueValues.get(j))});
                  }
                }
              }
              // now merge with the existing y ranges from the situation for this axis
              ArrayList existingYRanges = (ArrayList) situation.axisYPoints.get(axis.label);
              boolean[] axisYPixels = new boolean[axis.height + 1];
              for (int j = 0; j <= axis.height; j++) {
                for (int k = 0; k < yRanges.size(); k++) {
                  int[] yRange = (int[]) yRanges.get(k);
                  if (yRange.length == 2) {
                    if (axis.y + j >= yRange[0] && axis.y + j <= yRange[1]) {
                      axisYPixels[j] = true;
                    }
                  } else {
                    if (axis.y + j == yRange[0]) {
                      axisYPixels[j] = true;
                    }
                  }
                }
                for (int k = 0; k < existingYRanges.size(); k++) {
                  int[] yRange = (int[]) existingYRanges.get(k);
                  if (yRange.length == 2) {
                    if (axis.y + j >= yRange[0] && axis.y + j <= yRange[1]) {
                      axisYPixels[j] = true;
                    }
                  } else {
                    if (axis.y + j == yRange[0]) {
                      axisYPixels[j] = true;
                    }
                  }
                } 
              }
              ArrayList combinedValueRanges = new ArrayList();
              inRange = false;
              rangeStart = -1;
              for (int j = 0; j < axisYPixels.length; j++) {
                if (j < (axisYPixels.length - 1) && axisYPixels[j] && axisYPixels[(j + 1)]) { // sequential
                  if (!inRange) {
                    rangeStart = j;
                    inRange = true;
                  }
                } else {
                  if (inRange) {
                    inRange = false;
                    combinedValueRanges.add(new float[] {axis.getValueAt(axis.y + rangeStart), axis.getValueAt(axis.y + j)});
                  } else if (axisYPixels[j]) { // lone value
                    combinedValueRanges.add(new float[] {axis.getValueAt(axis.y + j)});
                  }
                }
              }
              situation.axisFloatValues.put(axis.label, combinedValueRanges);
            }
          }
          if (dataCasesSelections.size() > 0) {
            situation.calculatePoints(axisOrder, axes);
            situation.generateRangeBars(axisOrder, axes);
            updateBuffer = true;
          }
        }
      }
    }
  } else if (brushing) {
    brushing = false;
    Point brushEndPoint = new Point(mouseX, mouseY);
    boolean selections = false;
    for (int i = 0; i < dataCases.length; i++) {
      if (dataCases[i].intersectedBy(brushStartPoint, brushEndPoint)) {
        dataCases[i].colour = brushColour;
        dataCases[i].drawOrder++;
        selections = true;
      }
    }
    if (selections) updateBuffer = true;
  } else if (draggingAxis) {
    draggingAxis = false;
    int sourceAxisIndex = -1;
    int targetAxisIndex = -1;
    for (int i = 0; i < axisOrder.length; i++) {
      Axis axis = (Axis) axes.get(axisOrder[i]);
      if (axis.grabbed) {
        axis.grabbed = false;
        sourceAxisIndex = i;
      }
      if (axis.grabHandleHover()) {
        targetAxisIndex = i;
      }
    }
    if (sourceAxisIndex >= 0 && targetAxisIndex >= 0 && sourceAxisIndex != targetAxisIndex) {
      String targetAxisLabel = axisOrder[targetAxisIndex];
      axisOrder[targetAxisIndex] = axisOrder[sourceAxisIndex];
      axisOrder[sourceAxisIndex] = targetAxisLabel;
      windowResized();
    }
  }
}

void keyPressed() {
  if (addingSituation) {
    if (key == CODED && keyCode == CONTROL) {
      if (newSituationName.length() > 0) newSituationName.delete(0, newSituationName.length());
      addingSituation = false;
      updateBuffer = true;
    } else if (key == BACKSPACE) {
      if (newSituationName.length() > 0) newSituationName.deleteCharAt(newSituationName.length() - 1);
    } else if (key == RETURN || key == ENTER) {
      if (!situations.containsKey(newSituationName.toString()) && newSituationName.length() > 0) {
        Situation newSituation = new Situation(newSituationName.toString(), axes, axisOrder);
        situations.put(newSituation.label, newSituation);
        saveSituations();
        newSituationName.delete(0, newSituationName.length());
        addingSituation = false;
        calculateSituationPoints();
        updateBuffer = true;
      }
    } else if ((key >= 'a' && key <= 'z') || (key >= 'A' && key <= 'Z') || (key >= '0' && key <= '9') || key == ' ' || key == '(' || key == ')' || key == '-' || key == '+' || key == '&') {
      newSituationName.append(key);
    }
  } else {
    if (key == 'c') {
      for (int i = 0; i < dataCases.length; i++) { // for each data case
        dataCases[i].drawOrder = 0;
        dataCases[i].colour = dataCasesColour;
      }
      updateBuffer = true;
    }
  }
}

void saveSituations() {
  String[] situationStrings = new String[situations.size()];
  Set situationsLabels = situations.keySet();
  Iterator situationsLabelsIterator = situationsLabels.iterator();
  int i = 0;
  while (situationsLabelsIterator.hasNext()) {
    Situation situation = (Situation) situations.get((String) situationsLabelsIterator.next());
    situationStrings[i++] = situation.toString(numDimensions);
  }
  saveStrings("data/situations.tsv", situationStrings);
}
